//
//  FTForrstEngine.h
//  FTForrstEngine
//
//  Created by Benjamin Egelund-Müller (@bnmuller)

#import <Foundation/Foundation.h>

#import "FTComment.h"
#import "FTPost.h"
#import "FTUsageStatus.h"
#import "FTUser.h"

/**********************************************
 ** FTForrstEngine Constants 
 ***********************************************/

/* Block used for returning data from calls. 
 * 'response' and 'error' are never set at the same time. */
typedef void (^FTResponseHandler) (id response, NSError *error);

/* The error domain for any errors produced by the Forrst class. */
extern NSString * const FTErrorDomain;

/* Possible error codes for errors in FTErrorDomain. */
typedef enum {
	FTErrorCodeCredentialsNotSet,
	FTErrorCodeNoValuesSpecified,
	FTErrorCodeInvalidPostType,
	FTErrorCodeUserPostsLimitInvalid,
	FTErrorCodeServerError
} FTErrorCode;

/* Typedef NSString to FTPostSortType for more readable code. */
typedef NSString* FTPostSortType;

/* Possible post sorting methods. */
extern NSString * const FTPostSortTypeRecent;
extern NSString * const FTPostSortTypePopular;
extern NSString * const FTPostSortTypeBest;

/* Endpoint of the Forrst API. Currently points to forrst.com/api/v2/ */
extern NSString * const FTAPIEndpoint;

/* Constants for the user posts limit */
extern NSUInteger const FTUserPostsLimitMaximum;
extern NSUInteger const FTUserPostsLimitDefault;

/**********************************************
 ** FTForrstEngine Interface 
 ***********************************************/
@interface FTForrstEngine : NSObject {
	NSString *username; 
	NSString *password;
	NSString *accessToken;
	NSString *userAgent;
	BOOL useSSL; 
}

/* Can be either the username or email of the user */
@property (copy) NSString *username; 
@property (copy) NSString *password;

/* The access token will be set following authentication. */
@property (copy, readonly) NSString *accessToken;

/* User agent of the app. This is passed in the HTTP headers to uniquely identify your app. */
@property (copy) NSString *userAgent;

/* Defaults to NO. SSL is currently not supported by Forrst, but will be in the near future (supposedly). */
@property () BOOL useSSL; 

/* Calls the standard init method, then sets 'userAgent' according to the passed-in parameter. */
- (id) initWithUserAgent:(NSString *)aUserAgent;

/* Authenticates the user with the appropriate ivars and sets 'accessToken'. */
- (void) loginWithResponseHandler:(FTResponseHandler)aResponseHandler;

/* Sets the appropriate ivars, then calls -loginWithResponseHandler: */
- (void) loginWithUsername:(NSString *)aUsername password:(NSString *)aPassword responseHandler:(FTResponseHandler)aResponseHandler;

/* Loads the API usage stats. */
- (void) loadUsageStatsWithResponseHandler:(FTResponseHandler)aResponseHandler;

/* Loads information of the specified user. Note: 'aUser' need not contain all info. Only either 'userID' or 'username' needs be set. */
- (void) loadUserInfoWithUser:(FTUser *)aUser responseHandler:(FTResponseHandler)aResponseHandler;

/* Loads the posts of the specified user, of the specified post type. 
 * If limit is 0, it defaults to FTUserPostsLimitDefault. FTUserPostsLimitMaximum is the maximum. 
 * If 'aPost' is specified, only posts older than 'aPost' are returned. Passing nil will return the newest posts. */
- (void) loadUserPostsWithUser:(FTUser *)aUser postType:(FTPostType)aPostType limit:(NSUInteger)aLimit afterPost:(FTPost *)aPost responseHandler:(FTResponseHandler)aResponseHandler;

/* Loads all information about the specified post. Only the 'id' or 'tinyID' member variables needs be set. */
- (void) loadPostInformationWithPost:(FTPost *)aPost responseHandler:(FTResponseHandler)aResponseHandler;

/* Loads all posts after the specified post. If no post is specified, it will load the newest posts. */
- (void) loadAllPostsAfterPost:(FTPost *)aPost responseHandler:(FTResponseHandler)aResponseHandler;

/* Loads all posts meeting the specified parameters. Only postType is mandatory. Page defaults to 1 (according to the API), increase this to receive the following posts. */
- (void) loadPostListWithPostsOfType:(FTPostType)aPostType sortingType:(FTPostSortType)aSortType page:(NSUInteger)aPage responseHandler:(FTResponseHandler)aResponseHandler;

/* Loads the comments related to the specified post. Authentication required. */
- (void) loadCommentsForPost:(FTPost *)aPost responseHandler:(FTResponseHandler)aResponseHandler;

@end
