//
//  FTForrstEngine.m
//  FTForrstEngine
//
//  Created by Benjamin on 05/06/11.
//  Copyright 2011 Benjamin Egelund-Müller. All rights reserved.
//

#import "FTForrstEngine.h"

#import "ASIHTTPRequest.h"
#import "ASIFormDataRequest.h"

#import "JSONKit.h"

NSString * const FTErrorDomain = @"FTErrorDomain";
NSString * const FTAPIEndpoint = @"forrst.com/api/v2/";

NSUInteger const FTUserPostsLimitMaximum = 25; // As specified in the documentation, 06/06/2011.
NSUInteger const FTUserPostsLimitDefault = 10; // As specified in the documentation, 06/06/2011.

/**********************************************
 ** FTForrstEngine Interface Extension
 ***********************************************/
@interface FTForrstEngine ()

/* Redeclare to readwrite for internal use. */
@property (copy, readwrite) NSString *accessToken;

/* Used for administering the asynchronous network requests. */
@property (retain) NSOperationQueue *operationQueue;

/* Computes the API URL for the given method. Query strings is an array of NSStrings to be passed as GET parameters. */
- (NSURL *) urlWithMethod:(NSString *)method queryStrings:(NSArray *)params;

/* Returns a URL query component string identifying the user, suitable for passing to the API.  */
- (NSString *) queryStringComponentForUser:(FTUser *)aUser error:(NSError **)error;

/* Parses a response string containing a collection of posts into an array of FTPost objects. */
- (NSArray *) parsePostsFromResponseString:(NSString *)aResponseString error:(NSError **)error;

/* Deserializes 'responseString' into a dictionary, checks whether the server returned an error, then passes 
 * back the 'resp' dictionary of the response. */
- (NSDictionary *) extractAndVerifyResponseDictionaryFromString:(NSString *)responseString error:(NSError **)error;

@end


/**********************************************
 ** FTForrstEngine Implementation 
 ***********************************************/
@implementation FTForrstEngine

@synthesize username;
@synthesize password;
@synthesize accessToken;
@synthesize userAgent;

@synthesize useSSL;

@synthesize operationQueue;

- (void) dealloc 
{
	[username release], username = nil;
	[password release], password = nil;
	[accessToken release], accessToken = nil;
	[userAgent release], userAgent = nil;
	
	[operationQueue release], operationQueue = nil;
	
	[super dealloc];
}

#pragma mark -
#pragma mark Initialization

- (id) init 
{
	self = [super init];
	if (self)
	{
		useSSL = NO;
		operationQueue = [[NSOperationQueue alloc] init];
	}
	
	return self;
}

- (id) initWithUserAgent:(NSString *)aUserAgent
{
	self = [self init];
	if (self)
	{
		userAgent = [aUserAgent copy];
	}
	
	return self;
}

#pragma mark -
#pragma mark Instance Methods

- (void) loginWithResponseHandler:(FTResponseHandler)aResponseHandler
{
	// Pass an error and return if the username or password isn't set
	if (![username length] || ![password length]) {
		NSError *error = [NSError errorWithDomain:FTErrorDomain code:FTErrorCodeCredentialsNotSet userInfo:nil];
		aResponseHandler(nil, error);
		return;
	}
	
	// The method for authentication is 'users/auth'
	NSURL *authURL = [self urlWithMethod:@"users/auth" queryStrings:nil];
	
	// ASIFormDataRequest is used due to requirement to use POST 
	__block ASIFormDataRequest *request = [ASIFormDataRequest requestWithURL:authURL];
	request.userAgent = userAgent;
	
	[request setPostValue:username forKey:@"email_or_username"];
	[request setPostValue:password forKey:@"password"];
	
	[request setCompletionBlock:^(void) {
		// Get the responseDictionary and look for errors
		NSError *error = nil;
		NSDictionary *responseDictionary = [self extractAndVerifyResponseDictionaryFromString:[request responseString] error:&error];
		if (error) {
			// There was an error, return.
			aResponseHandler(nil, error);
			return;
		}
		
		NSString *token = [responseDictionary objectForKey:@"token"];
		
		self.accessToken = token;
		
		aResponseHandler((id)token, nil);
	}];
	[request setFailedBlock:^(void) {
		NSError *error = [request error];
		
		aResponseHandler(nil, error);
	}];
	
	// Don't start it directly, add it to the operation queue
	[operationQueue addOperation:request];
}

- (void) loginWithUsername:(NSString *)aUsername password:(NSString *)aPassword responseHandler:(FTResponseHandler)aResponseHandler
{
	self.username = aUsername;
	self.password = aPassword;
	[self loginWithResponseHandler:aResponseHandler];
}

- (void) loadUsageStatsWithResponseHandler:(FTResponseHandler)aResponseHandler
{
	NSURL *statsURL = [self urlWithMethod:@"stats" queryStrings:nil];
	
	__block ASIHTTPRequest *request = [ASIHTTPRequest requestWithURL:statsURL];
	request.userAgent = userAgent;
	
	[request setCompletionBlock:^(void) {
		NSError *error = nil;
		NSDictionary *responseDictionary = [self extractAndVerifyResponseDictionaryFromString:[request responseString] error:&error];
		if (error) {
			// There was an error, return.
			aResponseHandler(nil, error);
			return;
		}
		
		FTUsageStatus *usageStatus = [[[FTUsageStatus alloc] initWithResponseDictionary:responseDictionary] autorelease];
		
		aResponseHandler(usageStatus, nil);
	}];
	[request setFailedBlock:^(void) {
		NSError *error = [request error];
		
		aResponseHandler(nil, error);
	}];
	
	[operationQueue addOperation:request];
}

- (void) loadUserInfoWithUser:(FTUser *)aUser responseHandler:(FTResponseHandler)aResponseHandler;
{
	// Create the GET string
	NSError *error = nil;
	NSString *userQueryString = [self queryStringComponentForUser:aUser error:&error];
	if (!userQueryString) {
		aResponseHandler(nil, error);
		return;
	}
	
	// Create the URL using the query string
	NSURL *userInfoURL = [self urlWithMethod:@"users/info" queryStrings:[NSArray arrayWithObject:userQueryString]];
	
	// Create the request
	__block ASIHTTPRequest *request = [ASIHTTPRequest requestWithURL:userInfoURL];
	request.userAgent = userAgent;
	
	[request setCompletionBlock:^(void) {
		// Get the responseDictionary and look for errors
		NSError *error = nil;
		NSDictionary *responseDictionary = [self extractAndVerifyResponseDictionaryFromString:[request responseString] error:&error];
		if (error) {
			// There was an error, return.
			aResponseHandler(nil, error);
			return;
		}
		
		// Create the user from the dictionary
		FTUser *user = [[[FTUser alloc] initWithResponseDictionary:responseDictionary] autorelease];
		
		// Pass it back
		aResponseHandler(user, nil);
	}];
	[request setFailedBlock:^(void) {
		NSError *error = [request error];
		
		aResponseHandler(nil, error);
	}];
	
	// Don't start it directly, add it to the operation queue
	[operationQueue addOperation:request];
}

- (void) loadUserPostsWithUser:(FTUser *)aUser postType:(FTPostType)aPostType limit:(NSUInteger)aLimit afterPost:(FTPost *)aPost responseHandler:(FTResponseHandler)aResponseHandler
{
	NSMutableArray *queryStrings = [[[NSMutableArray alloc] init] autorelease];
	
	// Create the user query string and append to query string. Bail out if an error occurs.
	NSError *error = nil;
	NSString *userQueryString = [self queryStringComponentForUser:aUser error:&error];
	if (!userQueryString) {
		aResponseHandler(nil, error);
		return;
	}
	[queryStrings addObject:userQueryString];
	
	// If the post type is a known, valid one, use it. Otherwise ignore.
	if ([FTPost isValidPostType:aPostType]) {
		[queryStrings addObject:[NSString stringWithFormat:@"type=%@", aPostType]];
	}
	
	// If the limit is valid, use it. Otherwise use the default.
	if (0 < aLimit <= FTUserPostsLimitMaximum) {
		[queryStrings addObject:[NSString stringWithFormat:@"limit=%i", aLimit]];
	} else {
		[queryStrings addObject:[NSString stringWithFormat:@"limit=%i", FTUserPostsLimitDefault]];
	}
	
	// If the post has an ID that is not 0, use it.
	if (aPost.postID) {
		[queryStrings addObject:[NSString stringWithFormat:@"after=%i", aPost.postID]];
	}
	
	// Create the URL
	NSURL *userPostsURL = [self urlWithMethod:@"user/posts" queryStrings:queryStrings];
	
	// Create the request
	__block ASIHTTPRequest *request = [ASIHTTPRequest requestWithURL:userPostsURL];
	request.userAgent = userAgent;
	
	[request setCompletionBlock:^(void) {
		// Extract the posts
		NSString *responseString = [request responseString];
		
		NSError *error = nil;
		NSArray *posts = [self parsePostsFromResponseString:responseString error:&error];
		if (error) {
			aResponseHandler(nil, error);
			return;
		}
		
		// Pass it back
		aResponseHandler(posts, nil);
	}];
	[request setFailedBlock:^(void) {
		NSError *error = [request error];
		
		aResponseHandler(nil, error);
	}];
	
	// Don't start it directly, add it to the operation queue
	[operationQueue addOperation:request];
}

- (void) loadPostInformationWithPost:(FTPost *)aPost responseHandler:(FTResponseHandler)aResponseHandler
{
	NSString *queryString = nil;
	if (aPost.postID) {
		queryString = [NSString stringWithFormat:@"id=%i", aPost.postID];
	} else if (aPost.tinyID) {
		queryString = [NSString stringWithFormat:@"tiny_id=%@", aPost.tinyID];
	} else {
		NSError *error = [NSError errorWithDomain:FTErrorDomain code:FTErrorCodeNoValuesSpecified userInfo:nil];
		aResponseHandler(nil, error);
		return;
	}
	
	NSURL *postInfoURL = [self urlWithMethod:@"posts/show" queryStrings:[NSArray arrayWithObject:queryString]];
	
	__block ASIHTTPRequest *request = [ASIHTTPRequest requestWithURL:postInfoURL];
	request.userAgent = userAgent;
	
	[request setCompletionBlock:^(void) {
		// Get the responseDictionary and look for errors
		NSError *error = nil;
		NSDictionary *responseDictionary = [self extractAndVerifyResponseDictionaryFromString:[request responseString] error:&error];
		if (error) {
			// There was an error, return.
			aResponseHandler(nil, error);
			return;
		}
		
		// Create the post
		FTPost *post = [[[FTPost alloc] initWithResponseDictionary:responseDictionary] autorelease];
		
		aResponseHandler(post, nil);
	}];
	[request setFailedBlock:^(void) {
		NSError *error = [request error];
		
		aResponseHandler(nil, error);
	}];
	
	[operationQueue addOperation:request];
}

- (void) loadAllPostsAfterPost:(FTPost *)aPost responseHandler:(FTResponseHandler)aResponseHandler
{	
	NSArray *queryStrings = nil;
	
	// If the post has an ID that is not 0, use it. 
	if (aPost.postID) {
		queryStrings = [NSArray arrayWithObject:[NSString stringWithFormat:@"after=%i", aPost.postID]];
	}
	
	// Create the URL
	NSURL *allPostsURL = [self urlWithMethod:@"posts/all" queryStrings:queryStrings];
	
	// Create the request
	__block ASIHTTPRequest *request = [ASIHTTPRequest requestWithURL:allPostsURL];
	request.userAgent = userAgent;
	
	[request setCompletionBlock:^(void) {
		// Extract the posts
		NSString *responseString = [request responseString];
		
		NSError *error = nil;
		NSArray *posts = [self parsePostsFromResponseString:responseString error:&error];
		if (error) {
			aResponseHandler(nil, error);
			return;
		}
		
		// Pass it back
		aResponseHandler(posts, nil);
	}];
	[request setFailedBlock:^(void) {
		NSError *error = [request error];
		
		aResponseHandler(nil, error);
	}];
	
	// Don't start it directly, add it to the operation queue
	[operationQueue addOperation:request];
}

- (void) loadPostListWithPostsOfType:(FTPostType)aPostType sortingType:(FTPostSortType)aSortType page:(NSUInteger)aPage responseHandler:(FTResponseHandler)aResponseHandler;
{
	NSMutableArray *queryStrings = nil;
	
	// If the post has an ID that is not 0, use it. 
	if ([FTPost isValidPostType:aPostType]) {
		[queryStrings addObject:[NSString stringWithFormat:@"post_type=%@", aPostType]];
	}
	
	if (aSortType) {
		[queryStrings addObject:[NSString stringWithFormat:@"sort=%@", aSortType]];
	}
	
	if (aPage) {
		[queryStrings addObject:[NSString stringWithFormat:@"page=%i", aPage]];
	}
	
	// Create the URL
	NSURL *postListURL = [self urlWithMethod:@"posts/list" queryStrings:queryStrings];
	
	// Create the request
	__block ASIHTTPRequest *request = [ASIHTTPRequest requestWithURL:postListURL];
	request.userAgent = userAgent;
	
	[request setCompletionBlock:^(void) {
		// Extract the posts
		NSString *responseString = [request responseString];
		
		NSError *error = nil;
		NSArray *posts = [self parsePostsFromResponseString:responseString error:&error];
		if (error) {
			aResponseHandler(nil, error);
			return;
		}
		
		// Pass it back
		aResponseHandler(posts, nil);
	}];
	[request setFailedBlock:^(void) {
		NSError *error = [request error];
		
		aResponseHandler(nil, error);
	}];
	
	// Don't start it directly, add it to the operation queue
	[operationQueue addOperation:request];
}

- (void) loadCommentsForPost:(FTPost *)aPost responseHandler:(FTResponseHandler)aResponseHandler
{
	NSString *queryString = nil;
	if (aPost.postID) {
		queryString = [NSString stringWithFormat:@"id=%i", aPost.postID];
	} else if (aPost.tinyID) {
		queryString = [NSString stringWithFormat:@"tiny_id=%@", aPost.tinyID];
	} else {
		NSError *error = [NSError errorWithDomain:FTErrorDomain code:FTErrorCodeNoValuesSpecified userInfo:nil];
		aResponseHandler(nil, error);
		return;
	}
	
	NSURL *commentsURL = [self urlWithMethod:@"post/comments" queryStrings:[NSArray arrayWithObject:queryString]];
	
	__block ASIHTTPRequest *request = [ASIHTTPRequest requestWithURL:commentsURL];
	request.userAgent = userAgent;
	
	[request setCompletionBlock:^(void) {
		// Get the responseDictionary and look for errors
		NSError *error = nil;
		NSDictionary *responseDictionary = [self extractAndVerifyResponseDictionaryFromString:[request responseString] error:&error];
		if (error) {
			// There was an error, return.
			aResponseHandler(nil, error);
			return;
		}
		
		// Get the "unformatted" comment
		NSArray *unformattedComments = [responseDictionary objectForKey:@"comments"];
		
		// Loop through them all and create FTComment objects for each
		NSMutableArray *comments = [NSMutableArray arrayWithCapacity:[unformattedComments count]];
		
		for (NSDictionary *unformattedComment in unformattedComments) {
			FTComment *comment = [[FTComment alloc] initWithCommentDictionary:unformattedComment];
			[comments addObject:comment];
			[comment release];
		}
		
		// Pass it back
		aResponseHandler(comments, nil);
	}];
	[request setFailedBlock:^(void) {
		NSError *error = [request error];
		
		aResponseHandler(nil, error);
	}];
	
	// Don't start it directly, add it to the operation queue
	[operationQueue addOperation:request];
}

#pragma mark -
#pragma mark Private Methods

- (NSURL *) urlWithMethod:(NSString *)method queryStrings:(NSArray *)params
{
	NSMutableString *urlString = [[NSMutableString alloc] init];
	
	if (useSSL) {
		[urlString appendString:@"https://"];
	} else {
		[urlString appendString:@"http://"];
	}
	[urlString appendString:FTAPIEndpoint];
	[urlString appendString:method];
	
	if (accessToken) {
		NSString *accessTokenQuery = [NSString stringWithFormat:@"access_token=%@", accessToken];
		params = [params arrayByAddingObject:accessTokenQuery];
	}
	
	if (params) {
		[urlString appendString: @"?"];
		for (NSString *param in params) {
			[urlString appendString:param];
			[urlString appendString:@"&"];
		}
	}
	
	NSURL *url = [NSURL URLWithString:urlString];
	[urlString release];
	
	return url;
}

- (NSString *) queryStringComponentForUser:(FTUser *)aUser error:(NSError **)error
{
	NSString *retVal = nil;
	if (aUser.userID != 0) {
		retVal = [NSString stringWithFormat:@"id=%i", aUser.userID];
	} else if ([aUser.username length]) {
		retVal = [NSString stringWithFormat:@"username=%@", aUser.username];
	} else {
		*error = [NSError errorWithDomain:FTErrorDomain code:FTErrorCodeNoValuesSpecified userInfo:nil];
	}
	
	return retVal;
}

- (NSArray *) parsePostsFromResponseString:(NSString *)aResponseString error:(NSError **)error
{
	// Get the responseDictionary and look for errors
	
	NSDictionary *responseDictionary = [self extractAndVerifyResponseDictionaryFromString:aResponseString error:error];
	if (*error) {
		// There was an error, return.
		return nil;
	}
	
	// Get the "unformatted" posts
	NSArray *unformattedPosts = [responseDictionary objectForKey:@"posts"];
	
	// Loop through all the "unformatted" posts and convert them into FTPost objects.
	NSMutableArray *posts = [NSMutableArray arrayWithCapacity:[unformattedPosts count]];
	for (NSDictionary *unformattedPost in unformattedPosts) {
		FTPost *formattedPost = [[FTPost alloc] initWithResponseDictionary:unformattedPost];
		[posts addObject:formattedPost];
		[formattedPost release];
	}
	
	return posts;
}

- (NSDictionary *) extractAndVerifyResponseDictionaryFromString:(NSString *)responseString error:(NSError **)error
{
	NSDictionary *responseDictionary = [responseString objectFromJSONString];
	
	NSString *status = [responseDictionary objectForKey:@"stat"];
	if ([status isEqualToString:@"fail"]) {
		NSString *errorDescription = [[responseDictionary objectForKey:@"resp"] objectForKey:@"error"];
		
		NSDictionary *userInfo = [NSDictionary dictionaryWithObjectsAndKeys:errorDescription, NSLocalizedDescriptionKey, nil];
		*error = [NSError errorWithDomain:FTErrorDomain code:FTErrorCodeServerError userInfo:userInfo];
		
		return nil;
	} 
	
	return [responseDictionary objectForKey:@"resp"];
}

@end
